package ru.dbotthepony.mc.otm.client.screen

import com.mojang.blaze3d.platform.InputConstants
import com.mojang.blaze3d.vertex.PoseStack
import it.unimi.dsi.fastutil.ints.Int2FloatAVLTreeMap
import it.unimi.dsi.fastutil.ints.Int2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.ints.Int2ObjectFunction
import net.minecraft.ChatFormatting
import net.minecraft.client.Minecraft
import net.minecraft.client.resources.sounds.SimpleSoundInstance
import net.minecraft.network.chat.Component
import net.minecraft.sounds.SoundEvents
import net.minecraft.world.entity.player.Inventory
import ru.dbotthepony.mc.otm.TranslatableComponent
import ru.dbotthepony.mc.otm.android.AndroidResearch
import ru.dbotthepony.mc.otm.android.AndroidResearchType
import ru.dbotthepony.mc.otm.capability.MatteryPlayerCapability
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.client.render.Widgets18
import ru.dbotthepony.mc.otm.client.render.drawColor
import ru.dbotthepony.mc.otm.client.render.drawLine
import ru.dbotthepony.mc.otm.client.render.drawRect
import ru.dbotthepony.mc.otm.client.screen.panels.*
import ru.dbotthepony.mc.otm.client.screen.widget.PowerGaugePanel
import ru.dbotthepony.mc.otm.core.RGBAColor
import ru.dbotthepony.mc.otm.ifPresentK
import ru.dbotthepony.mc.otm.menu.AndroidStationMenu
import ru.dbotthepony.mc.otm.network.MatteryPlayerNetworkChannel
import ru.dbotthepony.mc.otm.network.AndroidResearchRequestPacket
import ru.dbotthepony.mc.otm.registry.MRegistry
import java.util.*
import kotlin.collections.ArrayList

private fun exploreTree(research: AndroidResearchType<*>, seen: MutableSet<AndroidResearchType<*>>, result: MutableList<AndroidResearchType<*>>) {
	if (!seen.add(research)) {
		return
	}

	result.add(research)

	for (children in research.flatUnlocks) {
		exploreTree(children, seen, result)
	}
}

private fun findGraphs(): List<Pair<AndroidResearchType<*>, List<AndroidResearchType<*>>>> {
	val seen = HashSet<AndroidResearchType<*>>()
	val list = ArrayList<Pair<AndroidResearchType<*>, ArrayList<AndroidResearchType<*>>>>()

	for (research in MRegistry.ANDROID_RESEARCH) {
		if (research.flatPrerequisites.isEmpty()) {
			val tree = ArrayList<AndroidResearchType<*>>()

			exploreTree(research, seen, tree)

			if (tree.isNotEmpty()) {
				list.add(research to tree)
			}
		}
	}

	return list
}

private fun isTree(root: AndroidResearchType<*>): Boolean {
	if (root.flatPrerequisites.size > 1) {
		return false
	}

	for (child in root.flatUnlocks) {
		if (!isTree(child)) {
			return false
		}
	}

	return true
}

typealias LinePos = Pair<Pair<Float, Float>, Pair<Float, Float>>

private class Tree(val node: AndroidResearchType<*>) : Iterable<Tree> {
	val subtrees = ArrayList<Tree>()

	val height: Int
	val width: Int

	init {
		for (child in node.flatUnlocks) {
			val tree = Tree(child)
			subtrees.add(tree)
		}

		if (subtrees.isEmpty()) {
			height = 0
			width = 1
		} else {
			var height = 0
			var width = 0

			for (child in subtrees) {
				height = height.coerceAtLeast(child.height)
				width += child.width
			}

			this.height = height + 1
			this.width = width
		}

		if (subtrees.isNotEmpty()) {
			subtrees.sortWith { a, b ->
				return@sortWith b.height.compareTo(a.height)
			}
		}
	}

	override fun iterator(): Iterator<Tree> {
		if (subtrees.size < 2) {
			return subtrees.iterator()
		} else {
			return object : Iterator<Tree> {
				private var index = subtrees.size - 1
				private var backward = true

				override fun hasNext(): Boolean {
					return backward || index < subtrees.size - 1
				}

				override fun next(): Tree {
					val index = index

					if (backward) {
						this.index -= 2

						if (this.index == -1) {
							backward = false
							this.index = 0
						} else if (this.index == 0) {
							backward = false
						}
					} else {
						if (this.index == 0) {
							this.index = if (subtrees.size % 2 == 0) 2 else 1
						} else {
							this.index += 2
						}
					}

					return subtrees[index]
				}
			}
		}
	}

	fun put(
		rows: Int2ObjectFunction<EditablePanel>,
		left: Float,
		capability: MatteryPlayerCapability
	): Pair<AndroidResearchButton, Float> {
		val totalWidth = width * 24f

		val lines = ArrayList<LinePos>()
		val linesToResearch = IdentityHashMap<AndroidResearchType<*>, ArrayList<LinePos>>()

		val button = AndroidResearchButton(rows[node.researchTreeDepth], capability.getResearch(node), lines, linesToResearch)
		button.x = left + totalWidth / 2f - button.width / 2f

		@Suppress("name_shadowing")
		var left = left

		var busLine: LinePos? = null

		if (subtrees.size > 1) {
			val x = button.width / 2f
			val y = button.y + button.height

			busLine = (x to y) to
					(x to y + 3f)

			lines.add(busLine)
		}

		var minX = Float.MAX_VALUE
		var maxX = -Float.MAX_VALUE

		for (subtree in this) {
			val (btn, move) = subtree.put(rows, left, capability)
			left += move

			val line: LinePos

			if (subtrees.size == 1) {
				val x = (button.width / 2f)
				val y = button.y + button.height

				line = (x to y) to
						(x to y + 6f)
			} else {
				val x = btn.x + btn.width / 2f - button.x
				val y = button.y + button.height + 3f

				line = (x to y) to
						(x to y + 3f)

				minX = minX.coerceAtMost(x)
				maxX = maxX.coerceAtLeast(x)

				val highlightLine =
					((button.width / 2f) to y) to
					(x to y)

				linesToResearch.computeIfAbsent(subtree.node) { ArrayList() }.add(highlightLine)

				for (children in subtree.node.allUnlocks) {
					linesToResearch.computeIfAbsent(children) { ArrayList() }.add(highlightLine)
				}
			}

			lines.add(line)

			linesToResearch.computeIfAbsent(subtree.node) { ArrayList() }.add(line)

			for (children in subtree.node.allUnlocks) {
				linesToResearch.computeIfAbsent(children) { ArrayList() }.add(line)
			}

			if (busLine != null) {
				linesToResearch.computeIfAbsent(subtree.node) { ArrayList() }.add(busLine)

				for (children in subtree.node.allUnlocks) {
					linesToResearch.computeIfAbsent(children) { ArrayList() }.add(busLine)
				}
			}
		}

		if (subtrees.size > 1) {
			val y = button.y + button.height + 3f

			lines.add(
				(minX - 0.5f to y) to
				(maxX to y)
			)
		}

		return button to totalWidth
	}
}

private open class AndroidResearchButton(
	parent: EditablePanel,
	private val node: AndroidResearch,
	private val lines: List<LinePos>,
	private val highlightLines: Map<AndroidResearchType<*>, List<LinePos>>
) :EditablePanel(
	parent.screen,
	parent,
	0f,
	0f,
	AndroidStationScreen.BUTTON_SIZE.toFloat(),
	AndroidStationScreen.BUTTON_SIZE.toFloat()
) {

	init {
		setDockMargin(2f, 2f, 2f, 2f)
	}

	override fun onHovered() {
		super.onHovered()

		(screen as AndroidStationScreen).hoveredResearch = node
	}

	override fun onUnHovered() {
		super.onUnHovered()

		if ((screen as AndroidStationScreen).hoveredResearch == node) {
			screen.hoveredResearch = null
		}
	}

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		val hovered = (screen as AndroidStationScreen).hoveredResearch

		val isBlockedByHovered = hovered != null && node.type in hovered.type.allBlocking

		if (isBlockedByHovered) {
			AndroidStationScreen.CAN_NOT_BE_RESEARCHED.setSystemColor()

			drawRect(stack, 0f, 0f, width, height)
		}

		if (node.isResearched) {
			AndroidStationScreen.RESEARCHED.setSystemColor()
		} else if (node.canResearch) {
			AndroidStationScreen.CAN_BE_RESEARCHED.setSystemColor()
		} else {
			AndroidStationScreen.CAN_NOT_BE_RESEARCHED.setSystemColor()
		}

		val icon = node.skinIcon

		if (icon != null) {
			icon.render(stack, 0f, 0f, width, height)
		} else {
			drawRect(stack, 0f, 0f, width, height)
		}

		if (isBlockedByHovered) {
			RGBAColor.RED.setSystemColor()

			Widgets18.CROSS.render(stack)
		}

		val text = node.iconText

		if (text != null) {
			font.drawShadow(stack, text, width - font.width(text), height - font.lineHeight, -0x1)
		}

		drawColor = RGBAColor.WHITE
		RGBAColor.WHITE.setShaderColor()

		for (line in lines) {
			val (pos1, pos2) = line
			val (x1, y1) = pos1
			val (x2, y2) = pos2

			drawLine(stack, x1, y1, x2, y2, 0.5f)
		}

		hovered ?: return
		val lines = highlightLines[hovered.type] ?: return

		drawColor = RGBAColor.LIGHT_GREEN

		for (line in lines) {
			val (pos1, pos2) = line
			val (x1, y1) = pos1
			val (x2, y2) = pos2

			drawLine(stack, x1, y1, x2, y2, 0.5f)
		}
	}

	override fun mouseClickedInner(x: Double, y: Double, button: Int): Boolean {
		if (button == InputConstants.MOUSE_BUTTON_LEFT) {
			if (node.canResearch && !node.isResearched) {
				MatteryPlayerNetworkChannel.sendToServer(AndroidResearchRequestPacket(node.type))
			}

			minecraft.soundManager.play(SimpleSoundInstance.forUI(SoundEvents.UI_BUTTON_CLICK, 1.0f))
		}

		return true
	}

	override fun innerRenderTooltips(stack: PoseStack, mouse_x: Float, mouse_y: Float, flag: Float): Boolean {
		if (isHovered) {
			val list = ArrayList<Component>().also { it.addAll(node.screenTooltipLines) }

			if (node.isResearched) {
				list.add(TranslatableComponent("otm.android_station.research.researched").withStyle(ChatFormatting.DARK_AQUA))
			} else if (node.canResearch) {
				list.add(TranslatableComponent("otm.android_station.research.can_be_researched").withStyle(ChatFormatting.DARK_GREEN))
			} else {
				list.add(TranslatableComponent("otm.android_station.research.can_not_be_researched").withStyle(ChatFormatting.DARK_RED))
			}

			screen.renderComponentTooltip(stack, list, mouse_x.toInt(), mouse_y.toInt())
		}

		return isHovered
	}
}

class AndroidStationScreen constructor(p_97741_: AndroidStationMenu, p_97742_: Inventory, p_97743_: Component) :
	MatteryScreen<AndroidStationMenu>(p_97741_, p_97742_, p_97743_) {

	private var canvas: DraggableCanvasPanel? = null
	private var research: FramePanel? = null
	var hoveredResearch: AndroidResearch? = null

	private fun openResearchTree() {
		val window = minecraft!!.window

		research = FramePanel(this, null, 0f, 0f, window.guiScaledWidth * 0.8f, window.guiScaledHeight * 0.8f, TranslatableComponent("otm.gui.android_research"))

		val rows = Int2ObjectAVLTreeMap<EditablePanel>()

		canvas = object : DraggableCanvasPanel(this@AndroidStationScreen, research, width = (GRID_WIDTH * 22).toFloat(), height = 0f) {
			override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
				drawColor = RGBAColor.BLACK
				drawRect(stack, 0f, 0f, width, height)
			}
		}

		minecraft?.player?.getCapability(MatteryCapability.MATTERY_PLAYER)?.ifPresentK {
			var totalWidth = 0f

			for (graph in findGraphs()) {
				if (isTree(graph.first)) {
					val tree = Tree(graph.first)

					for (i in 0 .. tree.height) {
						rows.computeIfAbsent(i, Int2ObjectFunction {
							object : EditablePanel(this@AndroidStationScreen, canvas, 0f, it * 24f, 10000f, 22f) {
								override fun mouseClickedInner(x: Double, y: Double, button: Int) = false
								override fun mouseReleasedInner(x: Double, y: Double, button: Int) = false
								override fun mouseDraggedInner(x: Double, y: Double, button: Int, xDelta: Double, yDelta: Double) = false
							}
						})
					}

					totalWidth += tree.put(rows, totalWidth, it).second
				} else {
					val rowWidths = Int2FloatAVLTreeMap()

					for (research in graph.second) {
						val row = rows.computeIfAbsent(research.researchTreeDepth, Int2ObjectFunction {
							object : EditablePanel(this@AndroidStationScreen, canvas, 0f, it * 24f, 10000f, 22f) {
								override fun mouseClickedInner(x: Double, y: Double, button: Int) = false
								override fun mouseReleasedInner(x: Double, y: Double, button: Int) = false
								override fun mouseDraggedInner(x: Double, y: Double, button: Int, xDelta: Double, yDelta: Double) = false
							}
						})

						val button = AndroidResearchButton(row, it.getResearch(research), emptyList(), emptyMap())
						val width = rowWidths[research.researchTreeDepth]
						button.setPos(totalWidth + width, 0f)
						rowWidths[research.researchTreeDepth] += 22f
					}

					var maximal = 0f

					for (value in rowWidths.values) {
						maximal = maximal.coerceAtLeast(value)
					}

					totalWidth += maximal
				}
			}
		}

		canvas!!.dock = Dock.FILL

		val bottom = EditablePanel(this, research, 0f, 0f, 0f, 20f)
		val close = ButtonPanel(this, bottom, 0f, 0f, 90f, 20f, TranslatableComponent("otm.container.matter_panel.close"))

		bottom.dock = Dock.BOTTOM

		close.dock = Dock.LEFT
		close.bind { research!!.remove() }

		bottom.setDockMargin(0f, 0f, 4f, 4f)
		canvas!!.setDockMargin(4f, 4f, 4f, 4f)

		research!!.toScreenCenter()
		addPanel(research!!)
	}

	override fun resize(minecraft: Minecraft, p_96576_: Int, p_96577_: Int) {
		super.resize(minecraft, p_96576_, p_96577_)

		val research = research

		if (research != null) {
			val window = minecraft.window
			research.setSize(window.guiScaledWidth * 0.8f, window.guiScaledHeight * 0.8f)
			research.toScreenCenter()
		}
	}

	override fun makeMainFrame(): FramePanel {
		val frame = super.makeMainFrame()!!

		PowerGaugePanel(this, frame, menu.powerWidget, LEFT_MARGIN, GAUGE_TOP_WITH_SLOT)
		SlotPanel(this, frame, menu.batterySlot, LEFT_MARGIN, SLOT_TOP_UNDER_GAUGE)
		SlotPanel(this, frame, menu.androidBattery, 38f, 17f)

		val button = ButtonPanel(this, frame, 38f, 69f, 124f, 20f, TranslatableComponent("otm.gui.android_research"))
		button.bind(this::openResearchTree)

		return frame
	}

	companion object {
		const val GRID_WIDTH = 5
		const val BUTTON_SIZE = 18
		val RESEARCHED = RGBAColor(150, 150, 200)
		val CAN_BE_RESEARCHED = RGBAColor(150, 200, 150)
		val CAN_NOT_BE_RESEARCHED = RGBAColor(200, 150, 150)
	}
}