package ru.dbotthepony.mc.otm.client.screen.panels

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.client.Minecraft
import net.minecraft.client.gui.narration.NarratableEntry
import net.minecraft.client.gui.narration.NarratableEntry.NarrationPriority
import net.minecraft.client.gui.narration.NarrationElementOutput
import net.minecraft.client.resources.sounds.SimpleSoundInstance
import net.minecraft.network.chat.Component
import net.minecraft.sounds.SoundEvents
import org.lwjgl.opengl.GL30
import ru.dbotthepony.mc.otm.client.render.*
import ru.dbotthepony.mc.otm.client.screen.MatteryScreen

open class FramePanel(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	x: Float,
	y: Float,
	width: Float,
	height: Float,
	protected var title: Component
) : EditablePanel(screen, parent, x, y, width, height), NarratableEntry {
	constructor(screen: MatteryScreen<*>, width: Float, height: Float, title: Component) : this(screen, null, 0f, 0f, width, height, title)

	enum class Position(val width: Float, val height: Float, val active_width: Float, val active_height: Float) {
		TOP(28f, 28f, 28f, 32f),
		// TODO: а оно вообще нужно?
		LEFT(28f, 28f, 32f, 28f),
		RIGHT(28f, 28f, 32f, 28f),
		BOTTOM(28f, 28f, 28f, 32f);
	}

	open inner class FrameTabPanel(position: Position, on_open: Runnable? = null, on_close: Runnable? = null) :
		EditablePanel(
			this@FramePanel.screen,
			this@FramePanel,
			0f,
			0f,
			position.width,
			position.height
		) {

		val frameTabPosition: Position = position
		var isActive = false
		var initial = false

		var onOpen: Runnable? = on_open
		var onClose: Runnable? = on_close

		open fun bindOnOpen(value: Runnable?) {
			onOpen = value
		}

		open fun bindOnClose(value: Runnable?) {
			onClose = value
		}

		fun showHidePanels(input: List<EditablePanel>) {
			onOpen = Runnable {
				for (child in input) {
					child.visible = true
				}
			}

			onClose = Runnable {
				for (child in input) {
					child.visible = false
				}
			}
		}

		override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
			if (frameTabPosition == Position.TOP) {
				val width: Float
				val height: Float

				if (isActive) {
					width = frameTabPosition.active_width
					height = frameTabPosition.active_height
					window_background.render(stack, 2f, 2f, width - 4, height - 2)
				} else {
					width = frameTabPosition.width
					height = frameTabPosition.height
					tab_background.render(stack, 2f, 2f, width - 4, height - 2)
				}

				top_window_border.renderW(stack, 3f, 0f, width - 6)
				left_window_border.renderH(stack, 0f, 3f, (height - if (isActive) if (initial) 2 else 4 else 3))

				right_window_border.renderH(stack, width - right_window_border.w, 3f, (height - if (isActive) 4 else 3))

				top_left_window_corner.render(stack, 0f, 0f)
				top_right_window_corner.render(stack, width - top_right_window_corner.w, 0f)

				if (isActive) {
					if (!initial) {
						tab_left_connection.render(stack, 0f, height - tab_left_connection.h - 1)
					}

					tab_right_connection.render(
						stack,
						width - tab_right_connection.w,
						height - tab_left_connection.h - 1
					)
				}
			}
		}

		protected open fun onOpen() {
			onOpen?.run()
		}

		protected open fun onClose() {
			onClose?.run()
		}

		override fun mouseClickedInner(x: Double, y: Double, flag: Int): Boolean {
			if (isActive) {
				return true
			}

			for (tab in tabsTop) {
				if (tab !== this) {
					if (tab.isActive) {
						tab.onClose()
						tab.isActive = false
					}
				}
			}

			for (tab in tabsLeft) {
				if (tab !== this) {
					if (tab.isActive) {
						tab.onClose()
						tab.isActive = false
					}
				}
			}

			for (tab in tabsRight) {
				if (tab !== this) {
					if (tab.isActive) {
						tab.onClose()
						tab.isActive = false
					}
				}
			}

			for (tab in tabsBottom) {
				if (tab !== this) {
					if (tab.isActive) {
						tab.onClose()
						tab.isActive = false
					}
				}
			}

			isActive = true
			onOpen()
			Minecraft.getInstance().soundManager.play(SimpleSoundInstance.forUI(SoundEvents.UI_BUTTON_CLICK, 1.0f))

			return true
		}
	}

	protected val tabsTop = ArrayList<FrameTabPanel>()
	protected val tabsLeft = ArrayList<FrameTabPanel>()
	protected val tabsRight = ArrayList<FrameTabPanel>()
	protected val tabsBottom = ArrayList<FrameTabPanel>()

	@JvmOverloads
	fun addTab(position: Position, on_open: Runnable? = null, on_close: Runnable? = null): FrameTabPanel {
		val tab = FrameTabPanel(position, on_open, on_close)
		doAddTab(tab)
		return tab
	}

	protected fun doAddTab(tab: FrameTabPanel) {
		if (tabsTop.size == 0 && tabsLeft.size == 0 && tabsRight.size == 0 && tabsBottom.size == 0) {
			tab.isActive = true
		}

		when (tab.frameTabPosition) {
			Position.TOP -> tabsTop.add(tab)
			Position.LEFT -> tabsLeft.add(tab)
			Position.RIGHT -> tabsRight.add(tab)
			Position.BOTTOM -> tabsBottom.add(tab)
		}
	}

	override fun performLayout() {
		for (i in tabsTop.indices) {
			val tab = tabsTop[i]
			tab.setPos(i * Position.TOP.width, -Position.TOP.height)
			tab.initial = i == 0
		}

		super.performLayout()
	}

	protected var dragging = false

	init {
		setDockPadding(PADDING, PADDING_TOP, PADDING, PADDING)
	}

	override fun mouseClickedInner(x: Double, y: Double, button: Int): Boolean {
		val (_, y1) = screenToLocal(x, y)

		if (parent == null && y1 >= 0 && y1 <= 12) {
			trapMouseInput = true
			dragging = true
			return true
		}

		return super.mouseClickedInner(x, y, button)
	}

	override fun mouseReleasedInner(x: Double, y: Double, button: Int): Boolean {
		if (dragging) {
			trapMouseInput = false
			dragging = false
			return true
		}

		return super.mouseReleasedInner(x, y, button)
	}

	override fun mouseDraggedInner(x: Double, y: Double, button: Int, xDelta: Double, yDelta: Double): Boolean {
		if (dragging) {
			this.x += xDelta.toFloat()
			this.y += yDelta.toFloat()
		}

		return super.mouseDraggedInner(x, y, button, xDelta, yDelta)
	}

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		// background
		// RenderSystem.disableDepthTest();
		val bgWidth = width - 4
		val bgHeight = height - 4

		// background
		if (bgWidth > 0 && bgHeight > 0) {
			window_background.render(stack, 3f, 3f, bgWidth, bgHeight)
		}

		// borders
		left_window_border.render(stack, 0f, 4f, 3f, height - 8)
		right_window_border.renderRaw(stack, width - 3, 4f, 3f, height - 6)
		top_window_border.renderRaw(stack, 4f, 0f, width - 8, 3f)
		bottom_window_border.renderRaw(stack, 4f, height - 3, width - 8, 3f)

		// corners
		top_left_window_corner.renderRaw(stack)
		top_right_window_corner.renderRaw(stack, x = width - 6)
		bottom_left_window_corner.renderRaw(stack, y = height - 6)
		bottom_right_window_corner.renderRaw(stack, x = width - 6, y = height - 6)

		// RenderSystem.enableDepthTest();

		// title
		RenderSystem.depthFunc(GL30.GL_ALWAYS)
		screen.font.draw(stack, title, 8f, 5f, 4210752)
		RenderSystem.depthFunc(GL30.GL_ALWAYS)
	}

	override fun narrationPriority(): NarrationPriority {
		return NarrationPriority.NONE
	}

	override fun updateNarration(narrationElementOutput: NarrationElementOutput) {}

	fun toScreenCenter() {
		setPos((screen.width / 2 - (width / 2).toInt()).toFloat(), (screen.height / 2 - (height / 2).toInt()).toFloat())
	}

	companion object {
		const val PADDING = 8f
		const val PADDING_TOP = 14f

		val top_left_window_corner 				= WidgetLocation.WIDGETS.element(x = 18f, y = 0f, w = 6f, h = 6f)
		val top_right_window_corner 			= WidgetLocation.WIDGETS.element(x = 24f, y = 0f, w = 6f, h = 6f)
		val left_window_border 					= WidgetLocation.WIDGETS.element(x = 18f, y = 4f, w = 3f, h = 5f)
		val right_window_border 				= WidgetLocation.WIDGETS.element(x = 27f, y = 3f, w = 3f, h = 5f)
		val top_window_border 					= WidgetLocation.WIDGETS.element(x = 22f, y = 0f, w = 5f, h = 3f)
		val window_background 					= WidgetLocation.WIDGETS.element(x = 30f, y = 12f, w = 6f, h = 6f)
		val tab_right_connection 				= WidgetLocation.WIDGETS.element(x = 30f, y = 0f, w = 3f, h = 5f)
		val tab_left_connection 				= WidgetLocation.WIDGETS.element(x = 33f, y = 0f, w = 3f, h = 5f)
		val tab_background 						= WidgetLocation.WIDGETS.element(x = 30f, y = 6f, w = 6f, h = 6f)
		val bottom_left_window_corner 			= WidgetLocation.WIDGETS.element(x = 18f, y = 6f, w = 6f, h = 6f)
		val bottom_right_window_corner 			= WidgetLocation.WIDGETS.element(x = 24f, y = 6f, w = 6f, h = 6f)
		val bottom_window_border 				= WidgetLocation.WIDGETS.element(x = 21f, y = 9f, w = 5f, h = 3f)
	}
}
